/*
 * Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */
#include <math.h>
#include <grdriver.h>
#include <grx.h>
#include "video.h"
#include "proto.h"
#include "dither.h"


int startx, starty;

/* Range values for lum, cr, cb. */
int LUM_RANGE;
int CR_RANGE;
int CB_RANGE;

/* Array that remaps color numbers to actual pixel values used by X server. */

unsigned char pixel[256];

/* Arrays holding quantized value ranged for lum, cr, and cb. */

int *lum_values;
int *cr_values;
int *cb_values;

/* Declaration of global variable containing dither type. */

extern int ditherType;

int vgamode = -1;



/*
 *--------------------------------------------------------------
 *
 * InitColor --
 *
 *	Initialized lum, cr, and cb quantized range value arrays.
 *
 * Results: 
 *      None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

void
InitColor()
{
  int i;

  for (i=0; i<LUM_RANGE; i++) {
    lum_values[i] = ((i * 256) / (LUM_RANGE)) + (256/(LUM_RANGE*2));
  }

  for (i=0; i<CR_RANGE; i++) {
    cr_values[i] = ((i * 256) / (CR_RANGE)) + (256/(CR_RANGE*2));
  }

  for (i=0; i<CB_RANGE; i++) {
    cb_values[i] = ((i * 256) / (CB_RANGE)) + (256/(CB_RANGE*2));
  }

}


/*
 *--------------------------------------------------------------
 *
 * ConvertColor --
 *
 *	Given a l, cr, cb tuple, converts it to r,g,b.
 *
 * Results:
 *	r,g,b values returned in pointers passed as parameters.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

static void
ConvertColor(l, cr, cb, r, g, b)
     unsigned char l, cr, cb;
     unsigned char *r, *g, *b;
{
  double fl, fcr, fcb, fr, fg, fb;

  fl = (double) l;
  fcr =  ((double) cr) - 128.0;
  fcb =  ((double) cb) - 128.0;


  fr = fl + (1.40200 * fcb);
  fg = fl - (0.71414 * fcb) - (0.34414 * fcr);
  fb = fl + (1.77200 * fcr);

  if (fr < 0.0) fr = 0.0;
  else if (fr > 255.0) fr = 255.0;

  if (fg < 0.0) fg = 0.0;
  else if (fg > 255.0) fg = 255.0;

  if (fb < 0.0) fb = 0.0;
  else if (fb > 255.0) fb = 255.0;

  *r = (unsigned char) fr;
  *g = (unsigned char) fg;
  *b = (unsigned char) fb;

}


void ResizeDisplay(int w, int h)
{
	int width, height, colors, done;
	GR_DRIVER_MODE_ENTRY *gr, *txt;

	width = height = colors = done = 0;
	GrGetDriverModes( &txt, &gr);

	done = 0;
	while( gr->width != 0  || done == 0) {
		if ((gr->number_of_colors == GrNumColors())  &&
                    (gr->width > w) && (gr->height > h)) {
		    	width  = gr->width;
			height = gr->height;
			colors = gr->number_of_colors;
			done = 1;
		}
		gr++;
	}

	if (width < w || height < h) {
	    	printf("MPEG screen size too big.\n");
	    	GrSetMode(GR_default_text,0,0,0);
	    	exit(-1);
	}
	GrSetMode(GR_width_height_color_graphics, width, height, colors);
}


static void
restoretextmode() {

  GrSetMode(GR_default_text,0,0,0);

}

static void 
MakeWindow(name) 
char *name;
{
	int width, height, colors, done;
	GR_DRIVER_MODE_ENTRY *gr, *text;
  
	if (ditherType == NO_DITHER) return;

	width = height = colors = 0;
	GrGetDriverModes( &text, &gr);
  	if (ditherType == FULL_COLOR_DITHER) {
		/* first find highest colormode that could work */
		while( gr->width != 0 ) {
			if (gr->number_of_colors > colors) {
				width  = gr->width;
				height = gr->height;
				colors = gr->number_of_colors;
			}
			gr++;
		}
		GrSetMode(GR_width_height_color_graphics, width, height, colors);
		GrClearScreen(GrBlack());
	        return;
    
	} else {
		while( gr->width != 0 ) {
			if (gr->number_of_colors == 256) {
				width  = gr->width;
				height = gr->height;
				colors = gr->number_of_colors;
			}
			gr++;
		}
		GrSetMode(GR_width_height_color_graphics, width, height, colors);
		GrClearScreen(GrBlack());
	        return;
	}

  	printf("Cannot find suitable SVGA graphics mode for selected dithering.\n");
  	exit(-1);

}


/*
 *--------------------------------------------------------------
 *
 * InitDisplay --
 *
 *	Initialized display, sets up colormap, etc.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

void InitDisplay(name)
char *name;
{
/* For 256 color modes. */

	int ncolors = LUM_RANGE*CB_RANGE*CR_RANGE;
	int i, lum_num, cr_num, cb_num;
	int r, g, b;

  	if (ditherType == NO_DITHER) return;

  	MakeWindow(name);

	for (i=0; i<ncolors; i++) {

		lum_num = (i / (CR_RANGE*CB_RANGE))%LUM_RANGE;
		cr_num = (i / CB_RANGE)%CR_RANGE;
		cb_num = i % CB_RANGE;

		r = g = b = 0;
		ConvertColor(lum_values[lum_num], cr_values[cr_num], cb_values[cb_num], &r, &g, &b);

		if (i > 256) {
			printf("mpeg_vga: not enough colors.\n");
			GrSetMode(GR_default_text,0,0,0);
			exit(-1);
		}
		GrSetColor(i, r , g , b );
		pixel[i] = i;
	}

}


/*
 *--------------------------------------------------------------
 *
 * InitGrayDisplay --
 *
 *	Initialized display for gray scale dither.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

#define NUM_COLORS 256

void InitGrayDisplay(char *name) {
	int ncolors = NUM_COLORS;
	int r, g, b, i;

	MakeWindow(name);
	for (i=0; i<ncolors; i++) {
		r = i;g = i; b = i;
		GrSetColor(i, r / 4, g / 4, b / 4);
		pixel[i] = i;
	}
}


/*
 *--------------------------------------------------------------
 *
 * InitMonoDisplay --
 *
 *	Initialized display for monochrome dither.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

void InitMonoDisplay(name)
char *name;
{

  MakeWindow(name);

}


/*
 *--------------------------------------------------------------
 *
 * InitColorDisplay --
 *
 *	Initialized display for full color output.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

void InitColorDisplay(name)
char *name;
{

  MakeWindow(name);

}


/*
 *--------------------------------------------------------------
 *
 * ExecuteDisplay --
 *
 *	Actually displays display plane in previously created window.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

void
ExecuteDisplay(VidStream *vid_stream) {

	GrContext *CurrentFrame;
static	GrContext *Screen = NULL;
	char *data;
	int width, height;

	
  	totNumFrames++;
  	if (!quietFlag) {
    		fprintf (stderr, "%d\r", totNumFrames);
  	}

  	if (ditherType == NO_DITHER) return;

  	if ( Screen == NULL) {
  	  GrCreateSubContext(0, 0, GrScreenX(),GrScreenY(), NULL, Screen);
          startx =  (GrScreenX() - vid_stream->h_size) >> 1;
          starty =  (GrScreenY() - vid_stream->v_size) >> 1;
        }
        
  	if (!noDisplayFlag) {
		data = (char *) vid_stream->current->display;
		height = vid_stream->v_size;
		width = vid_stream->h_size;
		
		if (ditherType == Twox2_DITHER) {
			width<<=1;
                        height<<=1;
		        startx =  (GrScreenX() - width) >> 1;
          	        starty =  (GrScreenY() - height) >> 1;
			CurrentFrame = GrCreateContext(width, height,data, NULL);
		} else if (ditherType == FULL_COLOR_DITHER) {
		        startx =  (GrScreenX() - width) >> 1;
          	        starty =  (GrScreenY() - height) >> 1;
			CurrentFrame = GrCreateContext(width, height,data, NULL);
        	} else if (ditherType == MONO_DITHER || ditherType == MONO_THRESHOLD) {
			CurrentFrame = GrCreateContext(width, height,data, NULL);
        	} else {
			CurrentFrame = GrCreateContext(width, height,data, NULL);
		}

  		GrBitBlt(Screen, startx, starty, CurrentFrame, 0, 0, width, height, GrWRITE);

      }

}
